/**
 * @file
 *
 * @copyright Axiros GmbH 2019-2021, All rights reserved
 *
 * @brief USP agent handling function definitions
 */

#include "compiler.h"

#ifndef USP_AGENT_H_
#define USP_AGENT_H_

struct usp_engine;
struct Usp__Msg;


/**
 *  @brief Our private usp_agent structure
 */
struct usp_agent;


/**
 * @brief Allocates memory and initialised an USP agent
 *
 * @param[in] engine	the engine to assign to this agent
 *
 * @return a new USP agent object
 */
struct usp_agent *usp_agent_new (struct usp_engine *engine) ARGS_NONNULL MALLOC RETURNS_NONNULL;


/**
 * @brief Retrieve the engine used by this USP agent
 *
 * @param[in] agent	the agent to retrieve the engine from
 *
 * @return the engine associated with this agent
 */
struct usp_engine *usp_agent_get_engine (struct usp_agent *agent) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Retrieve the frontend associated with this USP agent
 *
 * @param[in] agent	the agent to retrieve the frontend from
 *
 * @return the frontend associated with this agent
 */
struct frontend *usp_agent_get_frontend (struct usp_agent *agent) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Retrieve the endpoint identifier on an initialised USP agent
 *
 * @param[in] agent	the agent to retrieve the endpoint identifier from
 *
 * @return the endpoint identifier
 */
char *usp_agent_get_endpoint_id (struct usp_agent *agent) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Handle the incoming message and create a reply
 *
 * @param[in] agent	the agent to handle the message
 * @param[in] msg_in	the received USP Msg extracted from the USP Record
 * @param[in] from	the the sender of the USP Record
 *
 * @return a freshly allocated and appropriate reply message to send back or NULL in case no reply is necessary
 */
struct Usp__Msg *usp_agent_handle_message (struct usp_agent *agent, struct Usp__Msg *msg_in, char *from) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(3) MALLOC;


/**
 * @brief Free memory allocated by the USP agent object
 *
 * @param[in] agent	the USP agent to deconstruct
 */
void usp_agent_free (struct usp_agent *agent) ARGS_NONNULL;

#endif /* USP_AGENT_H_ */
