/**
 * @file
 *
 * @copyright Axiros GmbH 2015-2021, All rights reserved
 *
 * @brief Definition of datamodel object access and handling functions
 */

#ifndef DMOBJECT_H
#define DMOBJECT_H

#include <stddef.h>

#include "compiler.h"

#include "datamodel/DataModel.h"
#include "dmpermission.h"
#include "dmparameter.h"


/* Forward declarations */
struct dm_object;
struct dm_object_template;


/**
 * @brief Return the name of the object
 *
 * @param[in] obj The object to return the name for
 *
 * @return pointer to the name
 */
char *dm_object_get_name (struct dm_object *obj) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Return the permission matrix of the object
 *
 * @param[in] obj The object to return the permissions for
 *
 * @return permission matrix
 */
dm_permission_t dm_object_get_permissions (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Set the permission matrix of the object
 *
 * @param[in] obj The object to set the permissions for
 * @param[in] permissions The permission matrix to set on the object
 */
void dm_object_set_permissions (struct dm_object *obj, dm_permission_t permissions) ARGS_NONNULL;


/**
 * @brief Returns the parameter associated with this object, e.g. NumberOfEntries parameter for a table
 *
 * @param[in] table	The table object to return the associated parameter for
 *
 * @return The parameter associated with this object if if exists, NULL otherwise
 */
struct dm_parameter *dm_object_get_associated_parameter (struct dm_object *table) ARGS_NONNULL;


/**
 * @brief Extracts the instantiated path name of the object into the specified buffer starting at the specified object
 *
 * @param[in] obj	The object to return the relative path name for
 * @param[in] from	The parent of the object to include in the path, for the datamodel root this would be NULL
 * @param[in,out] buf	Buffer to store path in
 * @param[in] buflen	Size of buffer
 *
 * @note When obj == from, the path will be empty
 */
void dm_object_extract_relative_path (struct dm_object *obj, struct dm_object *from, char *buf, size_t buflen) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(3);


/**
 * @brief Extracts the full instantiated path name of the object into the specified buffer
 *
 * @param[in] obj The object to return the full path name for
 * @param[in,out] buf Buffer to store path in
 * @param[in] buflen Size of buffer
 */
void dm_object_extract_path (struct dm_object *obj, char *buf, size_t buflen) ARGS_NONNULL;


/**
 * @brief Extracts the supported path name of the object in TR-106 notation into the specified buffer starting at the specified object
 *
 * @param[in] obj	The object to return the relative path name for
 * @param[in] from	The parent of the object to include in the path, for the datamodel root this would be NULL
 * @param[in,out] buf	Buffer to store path in
 * @param[in] buflen	Size of buffer
 *
 * @note When obj == from, the path will be empty
 */
void dm_object_extract_relative_dm_path (struct dm_object *obj, struct dm_object *from, char *buf, size_t buflen) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(3);


/**
 * @brief Extracts the full supported path name of the object in TR-106 notation into the specified buffer
 *
 * @param[in] obj The object to return the full path name for
 * @param[in,out] buf Buffer to store path in
 * @param[in] buflen Size of buffer
 */
void dm_object_extract_dm_path (struct dm_object *obj, char *buf, size_t buflen) ARGS_NONNULL;


/**
 * @brief Returns the parent object of the object
 *
 * @param[in] obj The object to return the parent object for
 *
 * @return pointer to the parent object, NULL if this is the root object
 */
struct dm_object *dm_object_get_parent (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns the next child object from the given object using the specified path
 *
 * @param[in] obj The object to search a matching child object to
 * @param[in] path The path to follow to find the child. This MUST be a valid
 *                 partial path and MAY contain more than just the next level
 *                 component and if so the rest will be ignored
 *
 * @return pointer to the found child object, NULL if no object could be found or path is invalid
 */
struct dm_object *dm_object_get_next_child_in_path (struct dm_object *obj, char *path) ARGS_NONNULL;


/**
 * @brief Returns the object found by traversing the tree from the given object using the specified path
 *
 * @param[in]		obj	The object to search a matching leaf object from
 * @param[in,out]	path	A pointer to the path to follow to find the
 * 		  		leaf object. This MUST be a valid partial
 * 		  		object (regular, command or event) or parameter
 * 		  		path. The pointer will be advanced past the
 * 		  		processed components in the partial path so for
 * 		  		a fully processed object path it will point to
 * 		  		the terminating \0, for a fully processed
 * 		  		parameter path it will point to the parameter
 * 		  		name within the object.
 *
 * @return pointer to the found child object or NULL if no object could be found or path is invalid
 */
struct dm_object *dm_object_get_subobject_by_path (struct dm_object *obj, char **path) ARGS_NONNULL;


/**
 * @brief Returns the object found by traversing the datamodel from the root object using the specified full datamodel path
 *
 * @param[in]	obj	The root object (or any other object which will be used to find the real root)
 * @param[in]	path	A pointer to a fully qualified object or parameter path
 *
 * @return pointer to the found child object or NULL if no object could be found or path is invalid
 */
struct dm_object *dm_object_get_object_by_full_path (struct dm_object *obj, char *path) ARGS_NONNULL;


/**
 * @brief Returns the object found by traversing the tree from the given object using the specified path
 *
 * @param[in]	obj The object to search a matching object from
 * @param[in]	path A pointer to the path to follow to find the given concrete object
 *
 * @return pointer to the found object or NULL if no object could be found or the path is invalid
 */
struct dm_object *dm_object_get_object_by_path (struct dm_object *obj, char *path) ARGS_NONNULL;


/**
 * @brief Returns the object found by traversing the tree from the given object using the specified absolute path reference
 *
 * @param[in]	obj	The object used to find the datamodel root object
 * @param[in]	path	Path reference to the object (without trailing dot)
 *
 * @return pointer to the found object or NULL if no object could be found or the path reference is invalid
 */
struct dm_object *dm_object_get_object_by_reference (struct dm_object *obj, char *path) ARGS_NONNULL;


/**
 * @brief Returns the object found by traversing the tree from the given object using the specified absolute path reference
 *
 * @param[in]	obj	The object used to find the datamodel root object
 * @param[in]	path	Path reference to the object
 * @param[in]   len Path length to work with (without trailing dot)
 *
 * @return pointer to the found object or NULL if no object could be found or the path reference is invalid
 */
struct dm_object *dm_object_get_object_by_reference_len (struct dm_object *obj, char *path, size_t len) ARGS_NONNULL;


/**
 * @brief Returns the parameter found by traversing the tree from the given object using the specified path
 *
 * @param[in]	obj The object to search a matching parameter from
 * @param[in]	path A pointer to the path to follow to find the given concrete parameter
 *
 * @return pointer to the found parameter or NULL if no parameter could be found or the path is invalid
 */
struct dm_parameter *dm_object_get_parameter_by_path (struct dm_object *obj, char *path) ARGS_NONNULL;


/**
 * @brief Returns the parameter found by traversing the datamodel from the root object using the specified full datamodel path
 *
 * @param[in]	obj	The root object (or any other object which will be used to find the real root)
 * @param[in]	path	A pointer to a fully parameter path
 *
 * @return pointer to the found parameter object or NULL if no parameter could be found or path is invalid
 */
struct dm_parameter *dm_object_get_parameter_by_full_path (struct dm_object *obj, char *path) ARGS_NONNULL;


/**
 * @brief Returns whether the object is has the candidate as an ancestor
 *
 * @param[in]	obj The object to test for a relationship
 * @param[in]	candidate The candidate to check for ancestry
 *
 * @return 1 if the candidate is a direct ancestor of the obj
 */
unsigned int dm_object_is_ancestor (struct dm_object *obj, const struct dm_object *candidate) ARGS_NONNULL;


/**
 * @brief Returns whether the object is a table and allows adding instances
 *
 * @param[in] obj The object to check for creatability
 *
 * @return 1 if it is creatable, 0 otherwise
 */
unsigned int dm_object_is_creatable (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the object instance is deletable
 *
 * @param[in] obj The object to check for deletability
 *
 * @return 1 if it is a deletable instance, 0 otherwise
 */
unsigned int dm_object_is_deletable (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the object is marked as modified
 *
 * @param[in] obj The object to check for modified status
 *
 * @return 1 if it was modified, 0 otherwise
 */
unsigned int dm_object_is_modified (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Sets the modified status of this object
 *
 * @param[in] obj The object to set the modified status for
 */
void dm_object_set_modified (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Resets the modified status of this object
 *
 * @param[in] obj The object to reset the modified status for
 */
void dm_object_reset_modified (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Resets the updated status of this object
 *
 * @param[in] obj The object to reset the updated status for
 */
void dm_object_reset_updated (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the "object" is actually a USP command structure in disguise
 *
 * @param[in] obj The object to check for a command type
 *
 * @return 1 if it is an command, 0 otherwise
 */
unsigned int dm_object_is_command (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the "object" is a async USP command
 *
 * @param[in] obj The object to check for a async command type
 *
 * @return 1 if it is an async command, 0 otherwise
 */
unsigned int dm_object_is_async_command (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns the first virtual object ressembling an USP command associated with the specified object
 *
 * @param[in] obj The object to return the first command "object" for
 *
 * @return Pointer to the structure of the first command "object", or NULL if no commands are defined
 */
struct dm_object *dm_object_get_first_command (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Get the next sibling command relative to the specified command
 *
 * @param[in] command The object to get the next command "object" for
 *
 * @return Pointer to the structure of the next command "object", or NULL if no further commands are defined
 */
struct dm_object *dm_object_get_next_command (struct dm_object *command) ARGS_NONNULL;


/**
 * @brief Returns whether the object is actually an invisible data object
 *
 * @param[in] obj The object to check for a data type
 *
 * @return 1 if it is a data object, 0 otherwise
 */
unsigned int dm_object_is_data (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns the first virtual data object attached to the specified object
 *
 * @param[in] obj The object to return the first data object for
 *
 * @return Pointer to the structure of the data object, or NULL if no data objects are attached
 */
struct dm_object *dm_object_get_first_data (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Get the next sibling data object relative to the specified object
 *
 * @param[in] data The object to get the next data object for
 *
 * @return Pointer to the structure of the next data object, or NULL if no further data objects are attached
 */
struct dm_object *dm_object_get_next_data (struct dm_object *data) ARGS_NONNULL;


/**
 * @brief Allocate a new data object
 *
 * @param[in] name The name to use for the data object, may be NULL for anonymous data
 *
 * @return Pointer to a newly allocated (but not attached) data object
 */
struct dm_object *dm_object_create_data (char *name) RETURNS_NONNULL MALLOC;


/**
 * @brief Return whether the instance object is marked as deleted
 *
 * @param[in] obj	The instance object to return the deletion status for
 *
 * @return 1 if the object is currently marked as deleted, 0 otherwise
 */
unsigned int dm_object_is_deleted (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Mark the instance object as deleted. Use @ref dm_object_clear for other datamodel shaping.
 *
 * @param[in] obj	The instance object to mark as deleted
 */
void dm_object_delete (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Remove the deletion mark from an object
 *
 * @param[in] obj	The object to mark as not deleted
 */
void dm_object_undo_delete (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the "object" is actually a USP event structure in disguise
 *
 * @param[in] obj The object to check for a event type
 *
 * @return 1 if it is an event, 0 otherwise
 */
unsigned int dm_object_is_event (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns the first virtual object resembling an USP event associated with the specified object
 *
 * @param[in] obj The object to return the event "object" for
 *
 * @return Pointer to the structure of the event "object", or NULL if no events are defined
 */
struct dm_object *dm_object_get_first_event (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Get the next sibling event relative to the specified event
 *
 * @param[in] event The object to get the next event "object" for
 *
 * @return Pointer to the structure of the next event "object", or NULL if no further events are defined
 */
struct dm_object *dm_object_get_next_event (struct dm_object *event) ARGS_NONNULL;


/**
 * @brief Creates a new instance of an event and returns it
 *
 * @param[in] event	The event object to get the create a new instance for
 * @param[in] name	Unique event name
 *
 * @return Pointer to the created new event instance object or NULL if it failed
 */
struct dm_object *dm_object_create_event_instance (struct dm_object *event, char *name) ARGS_NONNULL;


/**
 * @brief Retrieves the first event instance from an event object
 *
 * @param[in] event	The event object to retrieve the first instance from
 *
 * @return Pointer to the first event instance object or NULL if there is none
 */
struct dm_object *dm_object_get_first_event_instance (struct dm_object *event) ARGS_NONNULL;


/**
 * @brief Retrieves the next event instance from an event instance
 *
 * @param[in] event	The event instance object to retrieve the next instance from
 *
 * @return Pointer to the next event instance object or NULL if there is none
 */
struct dm_object *dm_object_get_next_event_instance (struct dm_object *event) ARGS_NONNULL;


/**
 * @brief Returns whether the object is a table object
 *
 * @param[in] obj The object to check for a table type
 *
 * @return 1 if it is a table, 0 otherwise
 */
unsigned int dm_object_is_table (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the object is a in instance of a table
 *
 * @param[in] obj The object to check for a table instance type
 *
 * @return 1 if it is an instance of a table, 0 otherwise
 */
unsigned int dm_object_is_table_instance (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Set or clear the "don't backup" flag on an object
 *
 * @param[in] obj	The object to change the nobackup flag on
 * @param[in] flag	1 if the object shouldn't be backed up, 0 otherwise
 *
 * @note Setting this flag to 1 will also impact all parameters and child
 * 	 objects and cause their information not to be serialised. There're
 * 	 typically only very situations where this behaviour is desirable.
 */
void dm_object_set_nobackup (struct dm_object *obj, unsigned int flag) ARGS_NONNULL;

/**
 * @brief Returns whether the object is marked as "don't backup"
 *
 * @param[in] obj The object to check for being nobackup
 *
 * @return 1 if the object shouldn't be backed up, 0 otherwise
 */
unsigned int dm_object_is_nobackup (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns whether the object is a regular datamodel object
 *
 * @param[in] obj The object to check for being a regular object
 *
 * @return 1 if the object is a regular object in the TR-106 sense , 0 otherwise
 */
unsigned int dm_object_is_regular_object (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Create a new instance object from the template attached to the table
 *        object and link it in
 *
 * @param[in] table pointer to the table root to add the instance to
 * @param[in] no instance number to assign, use 0 for auto-assign
 *
 * @return pointer to the freshly created instance object or NULL in case of troubles
 *
 * @note If specified instance number already exists this function will fail
 *       and return NULL
 */
struct dm_object *dm_object_create_table_instance (struct dm_object *table, unsigned int no) ARGS_NONNULL;


/**
 * @brief Create a cloned object from the specified
 *
 * @param[in] tmpl pointer to object to be cloned
 *
 * @return pointer to the freshly created cloned object or NULL in case of troubles
 *
 * @note Please note that the cloned object will not have a name or parents
 */
struct dm_object *dm_object_clone (struct dm_object *tmpl) ARGS_NONNULL;


/**
 * @brief Set a custom index of a table instance
 *
 * @param[in] obj pointer to the table instance to set the name for
 * @param[in] no the number which to store in the instance
 *
 * @return returns 0 if renaming instance failed or 1 on success
 */
unsigned int dm_object_set_instance_number (struct dm_object *obj, unsigned int no) ARGS_NONNULL;

/**
 * @brief Get the index of a table instance
 *
 * @param[in] obj pointer to the table instance to get the number of
 *
 * @return returns 0 in case of a failure or the instance number in case of a success
 */
unsigned int dm_object_get_instance_number (struct dm_object *obj) ARGS_NONNULL;

/**
 * @brief Retrieve the instance with the given name
 *
 * @param[in] obj pointer to table root
 * @param[in] instance number or alias of the instance
 * @param[in] instlen length of the instance number or alias
 *
 * @return pointer to the instance or NULL if not found
 */
struct dm_object *dm_object_find_table_instance (struct dm_object *obj, char *instance, size_t instlen) ARGS_NONNULL;


/**
 * @brief Return the instance with the given number if it exists
 *
 * @param[in] obj	pointer to a table root
 * @param[in] no	the instance number to look for
 *
 * @return pointer to the instance or NULL if not found
 */
struct dm_object *dm_object_find_instance_by_number (struct dm_object *obj, unsigned int no) ARGS_NONNULL;


/**
 * @brief Count the number of instances below the table object
 *
 * @param[in] obj pointer to table root
 *
 * @return The number of instances found
 */
size_t dm_object_count_table_instances (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Find the child object with the given name
 *
 * @param[in] obj pointer to parent object
 * @param[in] name name of the object to look for
 *
 * @return pointer to object or NULL if not found
 *
 * @note This function will also find and return Event objects
 */
struct dm_object *dm_object_find_child_by_name (struct dm_object *obj, char *name) ARGS_NONNULL;


/**
 * @brief Find the child object with the given substring nam
 *
 * @param[in] obj pointer to parent object
 * @param[in] name name of the object to look for
 * @param[in] len length of the provided string to use
 *
 * @return pointer to object or NULL if not found
 *
 * @note This function will also find and return Event objects
 */
struct dm_object *dm_object_find_child_by_name_len (struct dm_object *obj, char *name, size_t len) ARGS_NONNULL;


/**
 * @brief Find the child object which has an Alias parameter identical to the given alias name
 *
 * @param[in] obj pointer to parent object
 * @param[in] alias alias name of the object to look for
 *
 * @return pointer to object or NULL if not found
 */
struct dm_object *dm_object_find_child_by_alias (struct dm_object *obj, char *alias) ARGS_NONNULL;


/**
 * @brief If the given object is a table object return the template for new instances
 *
 * @param[in] obj pointer to table root
 *
 * @return pointer to template object or NULL
 */
struct dm_object *dm_object_get_instance_template (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief clear out all data from the object
*
 * @param[in] obj pointer to the object to clear
 *
 * @return is either 0 for success or an TR-069 error code.
 */
unsigned int dm_object_clear (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns the internal metadata associated with the specified object
 *
 * @param[in] obj The object to return the internal metadata for
 *
 * @return pointer to char array specifying the internal metadata
 */
char *dm_object_get_intargs (struct dm_object *obj) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Append the specified parameters to the object
 *
 * @param[in] obj The object to append the parameter to
 * @param[in] par The parameter to append
 */
void dm_object_append_parameter (struct dm_object *obj, struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Set the internal metadata associated with an object to a copy of the specified string
 *
 * @param[in] obj The object to set the internal metadata for
 * @param[in] intargs The string with the metadata to put into the object
 */
void dm_object_set_intargs (struct dm_object *obj, char *intargs) ARGS_NONNULL;


/**
 * @brief Remove the internal metadata associated with an object
 *
 * @param[in] obj The object to delete the internal metadata from
 */
void dm_object_delete_intargs (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Create a new object from the specified template
 *
 * @param[in] tmpl The template object used to create the new object
 *
 * @return The allocated new object
 */
struct dm_object *dm_object_create_from_template (struct dm_object_template *tmpl) ARGS_NONNULL RETURNS_NONNULL MALLOC;


/**
 * @brief Create a new object from the specified template and add it to the datamodel as child of the given object if it does not exist
 *
 * @note This will (currently) not fuse parameters when adding another version of an already existing object!
 *
 * @param[in] obj The parent object to add the child to
 * @param[in] tmpl The template object used to create the new object
 *
 * @return The child object, either pre-existing or newly created
 */
struct dm_object *dm_object_add_from_template (struct dm_object *obj, struct dm_object_template *tmpl) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Returns the first parameter associated with the specified object
 *
 * @param[in] obj The object to return the first parameter for
 *
 * @return A parameter structure for the first parameter
 */
struct dm_parameter *dm_object_get_first_parameter (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Return the template used to create the object
 *
 * @param[in] obj The object to get the template for
 *
 * @return The template object used to create the object
 */
struct dm_object_template *dm_object_get_template (struct dm_object *obj) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Returns the number of parameters associated with the object
 *
 * @param[in] obj The object to return the number of parameters for
 *
 * @return Number of parameters in the object
 */
size_t dm_object_get_parameter_count (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Returns the total number of parameters associated with the object and any child objects
 *
 * @param[in] obj The object to recurively return the number of parameters for
 *
 * @return Number of parameters in the object or any child object
 */
size_t dm_object_get_recursive_parameter_count (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief find parameter structure in object by name
 *
 * @param[in] obj pointer to object
 * @param[in] paramname name of parameter
 *
 * @return pointer to parameter or NULL if not found
 */
struct dm_parameter *dm_object_get_parameter_by_name (struct dm_object *obj, char *paramname) ARGS_NONNULL;


/**
 * @brief parameter structure in object by name and return it's current value or NULL if the parameter couldn't be found
 *
 * @param[in] obj pointer to object
 * @param[in] paramname name of parameter
 *
 * @return the current value of the parameter or NULL
 */
char *dm_object_get_parameter_value (struct dm_object *obj, char *paramname) ARGS_NONNULL;


/**
 * @brief Set the the implicit parameter value of a parameter with the given
 * 	  name inside the specified object, turn off backup of the value and
 * 	  clear notification change flags
 *
 * @note This function is typically used to set up static parameter values during startup.
 *
 * @param[in] obj	pointer to the object to containing the parameter
 * @param[in] name	name of the parameter
 * @param[in] value	the value to set
 */
void dm_object_set_implicit_parameter_value (struct dm_object *obj, char *name, char *value) ARGS_NONNULL;


/**
 * @brief Set the value of a parameter with the given name inside the specified object
 *
 * @param[in] obj pointer to the object to containing the parameter
 * @param[in] name name of the parameter
 * @param[in] value the value to set
 * @param[in] check the level of checking to do on the parameter value
 *
 * @return returns 0 if changing the value was successful, 9005 if the
 * 	   parameter does not exist or any other TR-069 error value if
 * 	   the value couldn't be changed
 */
unsigned int dm_object_set_parameter_value (struct dm_object *obj, char *name, char *value, int check) ARGS_NONNULL;


/**
 * @brief Delete and free all previously allocated parameters from object
 *
 * @param[in] obj The object to remove all parameters from
 */
void dm_object_delete_parameters (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Output all information about the particular object to the system log
 *
 * @param[in] obj	The object to explain
 * @param[in] loglevel	The loglevel to use
 */
void dm_object_explain (struct dm_object *obj, int loglevel) ARGS_NONNULL;


/**
 * @brief Check recursively if there's any interest in serializing an object
 *
 * @param[in] obj pointer to object from which to start check
 *
 * @return 0 if there's no need for serialising, or 1 if there is
 */
unsigned int dm_object_check_save (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Frees the specified object and all associated data inside
 *
 * @param[in] obj The object to free
 */
void dm_object_free (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Allocate and set up a new dm_object structure
 *
 * @return The newly allocated object
 */
struct dm_object *dm_object_allocate (void) RETURNS_NONNULL MALLOC;


/**
 * @brief Append the initialised object to the specified parent object
 *
 * @param[in] parent The parent object to add the new child to
 * @param[in] child The child object to add to the end of the childrens of the parent
 * @return NULL if adding failed, child if the child was added as-is or (in case of a duplicate) the object that was extended with parameters from this object; the original child will be freed in this case
 *
 * @note You only need to care about the return value if you want to do further actions on the just added child, otherwise it is safe to ignore.
 */
struct dm_object *dm_object_add_child (struct dm_object *parent, struct dm_object *child) ARGS_NONNULL;


/**
 * @brief Unlink the specified child from its parent
 *
 * @param[in] child The child object to unlink from the childrens of the parent
 */
void dm_object_unlink_child (struct dm_object *child) ARGS_NONNULL;


/**
 * @brief Get the first interesting child object of the specified parent
 *
 * @param[in] obj The parent object to get the first child for
 *
 * @return Pointer to the first child object or NULL if parent has no children
 *
 * @note This will skip uninteresting objects like instance templates
 */
struct dm_object *dm_object_get_first_child (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Get the next sibling object of the specified obbject
 *
 * @param[in] obj The object to get the next sibling object for
 *
 * @return Pointer to the next sibling object or NULL
 */
struct dm_object *dm_object_get_next_sibling (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Get the object at the root of the tree of the specified object
 *
 * @param[in] obj The object to get the root object for
 *
 * @return Pointer to the root object of the tree
 */
struct dm_object *dm_object_get_root (struct dm_object *obj) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Get the frontend attached to this the datamodel the given object lives in
 *
 * @param[in] obj An object in the datamodel to get the frontend for
 *
 * @return Pointer to the frontend information
 */
struct frontend *dm_object_get_frontend (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Set the frontend for the datamodel the specified object lives in
 *
 * @param[in] obj	Any object in the datamodel to set the frontend for
 * @param[in] fe	The frontend to assign to the root object of datamodel
 */
void dm_object_set_frontend (struct dm_object *obj, struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Get the service object under which the specified object resides
 *
 * @param[in] obj The object to get the service object for
 *
 * @return Pointer to the service object of the the object resides in or NULL
 */
struct dm_object *dm_object_get_service (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Change the read or tabinit function for this object
 *
 * @param[in] obj The object to set the read or tabinit function on
 * @param[in] read_proc The read or tabinit function to use for this object
 *
 * @details This call will change the function in the template used for the
 *          particular object and hence will become active for all objects
 *          using the same template. Passing a NULL read procedure will disable
 *          the callback for this object.
 */
void dm_object_set_read_proc (struct dm_object *obj, dm_readfunc read_proc) ARGINDEX_NONNULL (1);


/**
 * @brief Change the write or tabmodity function for this object
 *
 * @param[in] obj The object to set the write or tabmodify function on
 * @param[in] write_proc The write or tabmodify function to use for this object
 *
 * @details This call will change the function in the template used for the
 *          particular object and hence will become active for all objects
 *          using the same template. Passing a NULL write procedure will
 *          disable the callback for this object.
 */
void dm_object_set_write_proc (struct dm_object *obj, dm_writefunc write_proc) ARGINDEX_NONNULL (1);


/**
 * @brief Call the read callback function for that object to update its content
 *        if it hasn't updated before in this session
 *
 * @param[in] obj The object call the read callback function for
 *
 * @return A CWMP error code
 */
unsigned int dm_object_update (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Call the read callback function for that object and all of its
 * 	  children to update their content if it hasn't updated before in this
 * 	  session
 *
 * @param[in] obj The object from which to start calling the read callback functions
 *
 * @return A CWMP error code
 */
unsigned int dm_object_update_tree (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Check wether any keys in this object instance conflict with any other object instances
 *
 * @param[in] obj The object instance to check the keys for
 *
 * @return 1 if the keys are unique or there are no keys, 0 otherwise
 */
unsigned int dm_object_has_valid_keys (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Retrieve the dm_read callback for the given object
 *
 * @param[in] obj The object to get the read callback function for
 *
 * @return The callable pointer for the read function
 */
dm_readfunc dm_object_get_read_proc (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Retrieve the dm_write callback for the given object
 *
 * @param[in] obj The object to get the write callback function for
 *
 * @return The callable pointer for the write function
 */
dm_writefunc dm_object_get_write_proc (struct dm_object *obj) ARGS_NONNULL;


/**
 * @brief Retrieve the dm_write callback for the given parameter
 *
 * @param[in] par The parameter to get the write callback function for
 *
 * @return The callable pointer for the write function
 *
 * @details If an overlay exists for the object containing this parameter the
 * 	    return value will be the dm_write function of the overlay
 * 	    (potentially also NULL) otherwise the dm_write function of the object
 */
dm_writefunc dm_object_get_parameter_write_proc (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Dump all information we have about the subobjects into the log output
 *
 * @param[in] root	the object to start dumping from
 * @param[in] loglevel	loglevel to use for dump
 */
void dm_object_dump_tree (struct dm_object *root, int loglevel) ARGS_NONNULL;


/**
 * @brief Type aware object's parameter value comparison to another value
 *
 * @param[in] obj The object to search a matching parameter from
 * @param[in] parname Parameter name
 * @param[in] value Parameter value
 *
 * @return enum dm_parameter_compare_result. If parameter is not found, DM_PC_NOT_SUPPORTED is returned.
 */
enum dm_parameter_compare_result dm_object_compare_parameter_value (struct dm_object *obj, char *parname, char *value) ARGS_NONNULL;


/**
 * @brief Type aware object's parameter value comparison to another value
 *
 * @param[in] obj The object to search a matching parameter from
 * @param[in] parname Parameter name
 * @param[in] value Parameter value
 * @param[in] len Parameter value length
 *
 * @return enum dm_parameter_compare_result. If parameter is not found, DM_PC_NOT_SUPPORTED is returned.
 */
enum dm_parameter_compare_result dm_object_compare_parameter_value_len (struct dm_object *obj, char *parname, char *value, size_t len) ARGS_NONNULL;


/**
 * @brief Get the next sibling object whose parameter (specified by name) semantically matches the value according to the criterion
 *
 * @param[in] obj The object to get the next sibling object for
 * @param[in] parname Parameter name
 * @param[in] value Parameter value
 * @param[in] crit Criterion
 *
 * @return Pointer to the next sibling object that matches the criterion or NULL
 */
struct dm_object *dm_object_get_next_sibling_by_crit (struct dm_object *obj, char *parname, char *value, enum dm_parameter_compare_result crit) ARGS_NONNULL;


/**
 * @brief Get the first child object whose parameter (specified by name) semantically matches the value according to the criterion
 *
 * @param[in] obj The object to get the next sibling object for
 * @param[in] parname Parameter name
 * @param[in] value Parameter value
 * @param[in] crit Criterion
 *
 * @return Pointer to the first child object that matches the criterion or NULL
 */
struct dm_object *dm_object_get_child_by_crit (struct dm_object *obj, char *parname, char *value, enum dm_parameter_compare_result crit) ARGS_NONNULL;


/**
 * @brief Retrieve the correct discriminant child according to the value of the discriminant parameter
 *
 * @note Currently there's no established mechanism for these union objects. At
 * 	 the moment this only works if the value of the discriminant parameter is
 * 	 equal to the name of the union member.
 *
 * @param[in] obj		The parent object containing the discriminant parameter and child object
 * @param[in] discparname	The parameter name of the discriminant parameter
 *
 * @return Pointer to the discriminant child according to the value of the given parameter or NULL if it doesn't exist
 */
struct dm_object *dm_object_get_child_by_discriminant (struct dm_object *obj, char *discparname) ARGS_NONNULL;


#endif /* DMOBJECT_H */
